@@ -12,9 +12,7 @@ relevant_files = [
     if f.endswith(".py") and not f in ["__init__.py"]
 ]
 
-hook_class_names = []
 connection_types = []
-name_template = "ewah.hooks.{0}.{1}"
 
 for py_file in relevant_files:
     mod = __import__(".".join([__name__, py_file]), fromlist=[py_file])
@@ -25,5 +23,9 @@ for py_file in relevant_files:
             and not cls == EWAHBaseHook
             and not cls == EWAHSQLBaseHook
         ):
-            hook_class_names.append(name_template.format(py_file, cls.__name__))
-            connection_types.append(cls.conn_type)
+            connection_types.append(
+                {
+                    "connection-type": cls.conn_type,
+                    "hook-class-name": ".".join([cls.__module__, cls.__name__]),
+                }
+            )
